/**************************************************************************
DESCRIPTION :
This module merges the commands and L1A from the TTC and VME:
- Any Hard reset generates also Soft_Rst.
- Start and Stop determine the state variable run_status;
- L1A is issued only when run_status=1;
- StRq is reset after a DAQ-frame transmission.
- GateTTCcommands inhibits TTC_Start and TTC_Stop; this will allow 
  the use of TTC with partial readout chains (Jeremiah Mans).
NB: DaqCntr relies on the fact that trigger are strictly 1-tick pulse.
****************************************************************************/

`include "define.v"
module commander  (
 input  clk,  L2IDLE_b, GateTTCcommands,	
        TTC_Hard_rst, TTC_Soft_rst, TTC_StRq, TTC_Resync,
        TTC_AsyncFifoRst, TTC_Start, TTC_Stop, TTC_L1A,  TTC_CalibTrig,
        VME_Hard_rst, VME_Soft_rst, VME_AsyncFifoRst,
        VME_Start, VME_Stop, VME_L1A, VME_CalibTrig, VME_StRq, 
 output reg AsyncFifoRst, L1A, CalibTrig,
            Hard_rst, Soft_rst, StRq, Start,
 output reg Stop = 1'b0, // without it, Stop powers up at 1 in ISE6.2.03
 output reg run_status = 1'b1         );
 
 reg VME_L1A_D,  TTC_L1A_D; // Delayed L1A's (for synchronization)
 reg VME_L1A_D2, TTC_L1A_D2; // Twice delayed L1A's
 reg VME_L1A_F,  TTC_L1A_F;  // Filtered L1A's
 reg OR_L1A;
 reg VME_CalibTrig_D,  TTC_CalibTrig_D; // Delayed CalibTrig's (for synchronization)
 reg VME_CalibTrig_D2, TTC_CalibTrig_D2; // Twice delayed CalibTrig's
 reg VME_CalibTrig_F,  TTC_CalibTrig_F;  // Filtered CalibTrig's
 reg OR_CalibTrig;
 
 reg L2transmission;	       

// ConfigRst asserted for ~3 clock cycles (tested), see also:
// www.xilinx.com/publications/xcellonline/xcell_51/xc_particle51.htm
// Note that FDS is set at power-up.
 FDS flop1 (.D(1'b0), .C(clk), .Q(out1), .S(1'b0)); 
 FD  flop2 (.D(out1), .C(clk), .Q(out2)); 
 FD  flop3 (.D(out2), .C(clk), .Q(out3)); 
 FD  flop4 (.D(out3), .C(clk), .Q(out4)); 
 assign ConfigRst = (out2 | out3 | out4); 

 always @ (posedge clk)  // no reset condition as this module generates the reset
   begin   // clk			 
    AsyncFifoRst <= VME_AsyncFifoRst | TTC_AsyncFifoRst;    
    Hard_rst <= TTC_Hard_rst | VME_Hard_rst | ConfigRst; // Hard_Rst implies Soft_rst...
    Soft_rst <= TTC_Soft_rst | VME_Soft_rst | TTC_Hard_rst | VME_Hard_rst | ConfigRst;
    Start <= VME_Start | (~GateTTCcommands & TTC_Start);
    Stop  <= VME_Stop  | (~GateTTCcommands & TTC_Stop);
    L2transmission <= L2IDLE_b; // delay to make sure daq_formatter latched StRq
    if (TTC_StRq | VME_StRq)  StRq  <= 1;
    if (L2transmission)       StRq  <= 0;
    if (Stop)   run_status <= 0; // Stop prevails on a simultaneous Start.
    else if (Start)  run_status <= 1;

    // edge-detection of all L1A's :
    TTC_L1A_D  <= TTC_L1A;     TTC_L1A_D2  <= TTC_L1A_D;  
    VME_L1A_D <= VME_L1A;    VME_L1A_D2 <= VME_L1A_D; 
    TTC_L1A_F  <= TTC_L1A_D  & ~TTC_L1A_D2;   
    VME_L1A_F <= VME_L1A_D & ~VME_L1A_D2;
    OR_L1A <= (TTC_L1A_F | VME_L1A_F);
    L1A <= OR_L1A & run_status;  // accept L1A only if run_status=1
    // edge-detection of all CalibTrig's :
    TTC_CalibTrig_D  <= TTC_CalibTrig;     TTC_CalibTrig_D2  <= TTC_CalibTrig_D;  
    VME_CalibTrig_D <= VME_CalibTrig;    VME_CalibTrig_D2 <= VME_CalibTrig_D; 
    TTC_CalibTrig_F  <= TTC_CalibTrig_D  & ~TTC_CalibTrig_D2;   
    VME_CalibTrig_F <= VME_CalibTrig_D & ~VME_CalibTrig_D2;
    OR_CalibTrig <= (TTC_CalibTrig_F | VME_CalibTrig_F);
    CalibTrig <= OR_CalibTrig & run_status;  // accept CalibTrig only if run_status=1
  end   // clk	     

endmodule  