/************************************************************************************
MODULE     :   LocBusSlave        AUTHOR    :   Tullio Grassi

DESCRIPTION :
The Local Bus interfaces different components on a board. This bus includes the data
lines, the address lines, WRITE_b (active low) and several CSn_b (active low
ChipSelect: CS1_b, CS2_b, for all the  slave devices).
The protocol foresees that when the master drives CS_b low for a particular slave,
the slave must decode the address on the Loc_Address lines, and read or write the 
Loc_Data bus depending of the WRITE_b direction.
On a write cycle, the master set WRITE_b = 0, set the address and data on the Local Bus,
and the slave will latch the data when CS_b = 0. 
The master keeps the data valid for the whole duration of the CS_b pulse. 
On a read cycle, the master set WRITE_b = 1, set the address on the Local Bus and
latch the data on the rising edge of CS_b, as the slave must prepare the data 
during the CS_b pulse. The CS_b pulse width depends on the slower slave.
A value of 300 ns is usually fine. When a slave is not selected, its data port is on Hi-Z.
The design is fully synchronous with clk. 
LocalBus lines are syncronized (this adds a few ticks of delay for edge detection).
It interfaces with FIFOs (Flow-Thru type) external to this verilog module.
**********************************************************************************/

`include "define.v"
module LocBusSlave   (      
// these ports connect to the LocalBus
  input  LocCS_b,  LocWRITE_b, // LOW = Write , HIGH = Read
  input  [`LB_Addr_Size-1:0]  LocAddr,  // The Address bus.
  inout  [`REG_WIDTH-1:0]     LocData,  // The bidir Data bus
// these ports connect to the rest of the FPGA
  input clk,  rst,
  input DLL_lock, event_buffer_empty, run_status, TTC_Ready, BunchEr, // to Status Reg
        L2OutSpyFIFOempty, L2OutSpyFIFOfull,      // to FIFO_StatReg
  input [2:0] UnlockCnt,
  input [3:0] AFifo1Status, AFifo2Status, AFifo3Status, AFifo4Status, 
              AFifo5Status, AFifo6Status, AFifo7Status, AFifo8Status,       
  input [7:0] LastTTCcommand, TTC_DbErrCnt, TTC_SinErrCnt,
  input [8:1] InSpyFIFO_empty, InSpyFIFO_full,   // to FIFO_StatReg
              SD, AFifoEmpty, AFifoFull, LW, FE_link_ERR, FE_link_DV,  
  input [15:0] TTC_test0, TTC_test1, TTC_test2, TTC_test3,
               TTC_test4, TTC_test5, TTC_test6, TTC_test7,
               TTC_test8, TTC_test9, TTC_testA, TTC_testB,
               TTC_testC, TTC_testD, TTC_testE, TTC_testF,
    ErrAcc1, ErrAcc2, ErrAcc3, ErrAcc4, ErrAcc5, ErrAcc6, ErrAcc7, ErrAcc8,
  input [17:0] Pattern1Back, Pattern2Back, Pattern3Back, Pattern4Back,
               Pattern5Back, Pattern6Back, Pattern7Back, Pattern8Back,
  input [18:0]  SnapFIFO1,  SnapFIFO2,  SnapFIFO3, SnapFIFO4,
                SnapFIFO5,  SnapFIFO6,  SnapFIFO7, SnapFIFO8,
  input [`EVN_WIDTH-1:0] EvN, CTrEvN,
  input [`SpyFIFO_WIDTH-1:0]   L2OutSpyFIFO_out,
           InSpyFIFO1_out,    InSpyFIFO2_out,
           InSpyFIFO3_out,    InSpyFIFO4_out,
           InSpyFIFO5_out,    InSpyFIFO6_out,
           InSpyFIFO7_out,    InSpyFIFO8_out,  
  input [`BCN_WIDTH-1:0]
      BCNofFib1BZero, BCNofFib2BZero, BCNofFib3BZero, BCNofFib4BZero, 
      BCNofFib5BZero, BCNofFib6BZero, BCNofFib7BZero, BCNofFib8BZero,
  // LUT stuff
  input [10:0]         
        InLUT_DOut1,  InLUT_DOut2,  InLUT_DOut3,  InLUT_DOut4,
        InLUT_DOut5,  InLUT_DOut6,  InLUT_DOut7,  InLUT_DOut8,
        InLUT_DOut9,  InLUT_DOut10, InLUT_DOut11, InLUT_DOut12,
        InLUT_DOut13, InLUT_DOut14, InLUT_DOut15, InLUT_DOut16,
        InLUT_DOut17, InLUT_DOut18, InLUT_DOut19, InLUT_DOut20,
        InLUT_DOut21, InLUT_DOut22, InLUT_DOut23, InLUT_DOut24,     
  input [9:0] ETSpy,
  input ETSpy_empty,ETSpy_full,
  output  reg  [24:1]    InLUT_WE,
  output  reg            InLUT_Mode,
  input [7:0]          
        OutLUT_DOut1,  OutLUT_DOut2,  OutLUT_DOut3,  OutLUT_DOut4,
        OutLUT_DOut5,  OutLUT_DOut6,  OutLUT_DOut7,  OutLUT_DOut8,
        OutLUT_DOut9,  OutLUT_DOut10, OutLUT_DOut11, OutLUT_DOut12,
        OutLUT_DOut13, OutLUT_DOut14, OutLUT_DOut15, OutLUT_DOut16,
        OutLUT_DOut17, OutLUT_DOut18, OutLUT_DOut19, OutLUT_DOut20,
        OutLUT_DOut21, OutLUT_DOut22, OutLUT_DOut23, OutLUT_DOut24,         
  output  reg  [24:1]   OutLUT_WE,
  output  reg           OutLUT_Mode,
  output  reg ETSpyFIFO_RD,
  output  reg [9:0] ETlow, EThigh,    // compare to the 10-bit lineared ET, for Luminosity
  output  reg [4:0] ETch_mux,         // determines which channel get muxed into the ET spy fifo

  output  ClockSelect, PatternMode, RepeatMode, LatTrim,
          CompactNorm, CompactCalib, GateTTCcommands, StartStop_Spy, // <=> Cntr Reg
          VME_Hard_rst, VME_Soft_rst, VME_StRq, VME_CalibTrig, VME_ResCTrEvN,
          VME_Start, VME_Stop, VME_L1A, VME_AsyncFifoRst, VME_TrigOut,
          TTC_Reset_b,  // reset pulse generated from a Local Bus write access
          bypass, CounterModeSel, 
  output  [8:1] TI_EN,   // individual enable for TLK2501
  output  [`REG_WIDTH-1:0] D_IN, // output for debug  
  output  reg Tri,   // output for test only  
  output  reg InSpyFIFO1_RD, InSpyFIFO2_RD,  InSpyFIFO3_RD,  InSpyFIFO4_RD,
              InSpyFIFO5_RD, InSpyFIFO6_RD,  InSpyFIFO7_RD,  InSpyFIFO8_RD,
  output  reg L2OutSpyFIFO_RD,
  output  reg [2:0] LB_STATE,       // output for debug  
  output  reg [4:0] NS, NPS, //Programmable # of Samples and Pre-Samples (per L1A)
  output  reg [8:1] PatternsWrEn, SnapRdEn,
  output  reg [15:0] ProgInputDelay,
  output  reg [17:0] 
          Threshold1,  Threshold2,  Threshold3,  Threshold4, 
          Threshold5,  Threshold6,  Threshold7,  Threshold8, 
          Threshold9,  Threshold10, Threshold11, Threshold12, 
          Threshold13, Threshold14, Threshold15, Threshold16, 
          Threshold17, Threshold18, Threshold19, Threshold20, 
          Threshold21, Threshold22, Threshold23, Threshold24, 
  output  reg [23:0] QieChMask,    // masks for TPG
  output  reg [`BCN_WIDTH-1:0]  MaxBCN ='hDEB, SetBCN,
  output  reg [`REG_WIDTH-1:0] D_OUT, HTRsubmodN, PipeLength       ); 

  parameter      //  FPGAversionN moved to define.v

  //  Parameters used for the STATEs of the LocBus FSM
  IDLE           =  0,  //  Waiting for the negative transition on LocCS_b
  RD_OR_WR       =  1,  //  check LocWRITE_b.  
  MASTER_WRITE   =  2,  //  If WR mode, then Store data
  MASTER_READ    =  3,  //  If RD mode, prepare data to be read
  END_OF_READ    =  4,  //  Latch data, release control lines and wait posedge of LocCS_b
  END_OF_WRITE   =  5,  //  Stop WR pulses and wait for posedge of LocCS_b
  END_CYCLE      =  6;  //  Set Data lines to HiZ (has to be after posedge)

//     INTERNAL SIGNALS
  wire TTC_Reset;
  wire [8:1] TI_EN_b;
  reg  [`REG_WIDTH-1:0] StatusReg, FIFO_StatReg, FE_StatReg, CntrReg;
  reg  [`REG_WIDTH-1:0] WriteOnlyCntrReg;
  reg  LocCS_Del_0, LocCS_Del_1, LocCS_Del_2;    // delayed versions of CS
  reg  [`BCN_WIDTH-1:0]
       BCNofFib1BZero_r, BCNofFib2BZero_r, BCNofFib3BZero_r, BCNofFib4BZero_r, 
       BCNofFib5BZero_r, BCNofFib6BZero_r, BCNofFib7BZero_r, BCNofFib8BZero_r;  
  
//  ACTUAL CODE STARTS HERE
//  assign  Histo_Mode         =  CntrReg[0];     RESERVED
//  assign  histType           =  CntrReg[1];
  assign  ClockSelect        =  CntrReg[2];
  assign  CompactNorm        =  CntrReg[3];
  assign  CompactCalib       =  CntrReg[4];
  assign  LatTrim            =  CntrReg[6];
  assign  RepeatMode         =  CntrReg[7];
  assign  PatternMode        =  CntrReg[8];
  assign  bypass             =  CntrReg[9];
  assign  CounterModeSel     =  CntrReg[10];
  assign  GateTTCcommands    =  CntrReg[11];
  assign  StartStop_Spy      =  CntrReg[12];

  //  WriteOnlyCntrReg self-reset itself and creates positive pulses
  assign  VME_Hard_rst        =   WriteOnlyCntrReg[0];
  assign  VME_Soft_rst        =   WriteOnlyCntrReg[1];
  assign  VME_Start           =   WriteOnlyCntrReg[2];
  assign  VME_Stop            =   WriteOnlyCntrReg[3];
  assign  VME_StRq            =   WriteOnlyCntrReg[4];
  assign  TTC_Reset           =   WriteOnlyCntrReg[5]; 
  assign  VME_L1A             =   WriteOnlyCntrReg[6]; 
  assign  VME_AsyncFifoRst    =   WriteOnlyCntrReg[7]; 
  assign  TI_EN_b[1]          =   WriteOnlyCntrReg[8]; 
  assign  TI_EN_b[2]          =   WriteOnlyCntrReg[9];
  assign  TI_EN_b[3]          =   WriteOnlyCntrReg[10];
  assign  TI_EN_b[4]          =   WriteOnlyCntrReg[11];
  assign  TI_EN_b[5]          =   WriteOnlyCntrReg[12];
  assign  TI_EN_b[6]          =   WriteOnlyCntrReg[13];
  assign  TI_EN_b[7]          =   WriteOnlyCntrReg[14];
  assign  TI_EN_b[8]          =   WriteOnlyCntrReg[15];
  assign  VME_TrigOut         =   WriteOnlyCntrReg[16]; 
  assign  VME_CalibTrig       =   WriteOnlyCntrReg[17];
  assign  VME_ResCTrEvN       =   WriteOnlyCntrReg[18]; 

  assign  TTC_Reset_b         =   ~TTC_Reset; // TTC_Reset_b must be a negative pulse
  assign  TI_EN               =   ~TI_EN_b;     // Negative pulses to reset TI deser.

// forcing IOBUFfer to convice XST.  LocData goes stright to the pads while the rest of
//  the logic makes use of D_IN and D_OUT. NB: T=1 means Tristate.
//    _S_2  means Slow output, 2 mA drive (minimum, in case of bus collisions)
 IOBUF_S_2   LocData_BUF0  (.IO(LocData[0]),  .I(D_OUT[0]),  .O(D_IN[0]),  .T(Tri));
 IOBUF_S_2   LocData_BUF1  (.IO(LocData[1]),  .I(D_OUT[1]),  .O(D_IN[1]),  .T(Tri));
 IOBUF_S_2   LocData_BUF2  (.IO(LocData[2]),  .I(D_OUT[2]),  .O(D_IN[2]),  .T(Tri));
 IOBUF_S_2   LocData_BUF3  (.IO(LocData[3]),  .I(D_OUT[3]),  .O(D_IN[3]),  .T(Tri));
 IOBUF_S_2   LocData_BUF4  (.IO(LocData[4]),  .I(D_OUT[4]),  .O(D_IN[4]),  .T(Tri));
 IOBUF_S_2   LocData_BUF5  (.IO(LocData[5]),  .I(D_OUT[5]),  .O(D_IN[5]),  .T(Tri));
 IOBUF_S_2   LocData_BUF6  (.IO(LocData[6]),  .I(D_OUT[6]),  .O(D_IN[6]),  .T(Tri));
 IOBUF_S_2   LocData_BUF7  (.IO(LocData[7]),  .I(D_OUT[7]),  .O(D_IN[7]),  .T(Tri));
 IOBUF_S_2   LocData_BUF8  (.IO(LocData[8]),  .I(D_OUT[8]),  .O(D_IN[8]),  .T(Tri));
 IOBUF_S_2   LocData_BUF9  (.IO(LocData[9]),  .I(D_OUT[9]),  .O(D_IN[9]),  .T(Tri));
 IOBUF_S_2   LocData_BUF10 (.IO(LocData[10]), .I(D_OUT[10]), .O(D_IN[10]), .T(Tri));
 IOBUF_S_2   LocData_BUF11 (.IO(LocData[11]), .I(D_OUT[11]), .O(D_IN[11]), .T(Tri));
 IOBUF_S_2   LocData_BUF12 (.IO(LocData[12]), .I(D_OUT[12]), .O(D_IN[12]), .T(Tri));
 IOBUF_S_2   LocData_BUF13 (.IO(LocData[13]), .I(D_OUT[13]), .O(D_IN[13]), .T(Tri));
 IOBUF_S_2   LocData_BUF14 (.IO(LocData[14]), .I(D_OUT[14]), .O(D_IN[14]), .T(Tri));
 IOBUF_S_2   LocData_BUF15 (.IO(LocData[15]), .I(D_OUT[15]), .O(D_IN[15]), .T(Tri));
 IOBUF_S_2   LocData_BUF16 (.IO(LocData[16]), .I(D_OUT[16]), .O(D_IN[16]), .T(Tri));
 IOBUF_S_2   LocData_BUF17 (.IO(LocData[17]), .I(D_OUT[17]), .O(D_IN[17]), .T(Tri));
 IOBUF_S_2   LocData_BUF18 (.IO(LocData[18]), .I(D_OUT[18]), .O(D_IN[18]), .T(Tri));

 always@(posedge clk)  begin  
   if(rst==1)   begin       //   RESET
         HTRsubmodN <=  'hCD;       PipeLength <=  'hC;
         NS <= 'd7; NPS <= 'd2;     ProgInputDelay <= 0;
			//
			// set the defaults for the energy thresholds....try 1 GeV and 10 GeV
			ETlow <= 'h1;    // assume 1 bit = 1 GeV
			EThigh <= 'h10;  // ditto
			ETch_mux <= 0;   // mux channel 0 by default (ch 0 out of 24, so 0-23)
   end  //   END of RESET 
   else  begin   // posedge clk       
      LocCS_Del_0  <=  LocCS_b;    // FF to avoid metastability
      LocCS_Del_1  <=  LocCS_Del_0;
      LocCS_Del_2  <=  LocCS_Del_1;  // total of 3Tck = 75 ns delay on CS
      StatusReg[0]                <=  DLL_lock;
      StatusReg[1]                <=  run_status;  
      StatusReg[7:2]              <=  LastTTCcommand[7:2];
      StatusReg[8]                <=  event_buffer_empty;
      StatusReg[9]                <=  TTC_Ready;
      StatusReg[12:10]            <=  UnlockCnt;
      StatusReg[13]               <=  BunchEr;
	   StatusReg[`REG_WIDTH-1:14]  <=  0;  // unused portion of status reg
      FIFO_StatReg[7:0]           <=  InSpyFIFO_empty[8:1];
      FIFO_StatReg[15:8]          <=  InSpyFIFO_full[8:1];
      FIFO_StatReg[16]            <=  L2OutSpyFIFOempty;
      FIFO_StatReg[17]            <=  L2OutSpyFIFOfull;
      FIFO_StatReg[18]            <=  0;
      FE_StatReg[7:0]             <=  FE_link_ERR[8:1]; 
      FE_StatReg[15:8]            <=  FE_link_DV[8:1];
      FE_StatReg[`REG_WIDTH-1:16] <=  0;  // unused portion of status reg
      BCNofFib1BZero_r  <=  BCNofFib1BZero;   BCNofFib2BZero_r  <=  BCNofFib2BZero;
      BCNofFib3BZero_r  <=  BCNofFib3BZero;   BCNofFib4BZero_r  <=  BCNofFib4BZero;
      BCNofFib5BZero_r  <=  BCNofFib5BZero;   BCNofFib6BZero_r  <=  BCNofFib6BZero;
      BCNofFib7BZero_r  <=  BCNofFib7BZero;   BCNofFib8BZero_r  <=  BCNofFib8BZero;
      
    case(LB_STATE)
        
    IDLE:  // check for a negative edge on LocCS_b (digitally filtered)
        begin
          D_OUT <=  'h BAD;    //mark a bad LocBus read timing
          Tri        <=  1;    // Data lines normally Hi-Z
          InLUT_WE   <=  0;    OutLUT_WE  <= 0;
          InLUT_Mode <= 0;     OutLUT_Mode <= 0;
          if(LocCS_Del_2 == 1 && LocCS_Del_1==0)  LB_STATE  <=  RD_OR_WR;
        end
            
    RD_OR_WR: //  decide whether it is a READ or a WRITE
        begin
          if (LocWRITE_b==0)
             LB_STATE <= MASTER_WRITE; // Write operation (master's point of view)
          else  begin 
             LB_STATE <= MASTER_READ;  // Read operation (master's point of view)
             Tri      <= 0;            // Set Data lines toward LocalBus
            end  //  else
        end          
      //  ***********************************************   
      //     ********   W R I T E  operation  ********    
    MASTER_WRITE:  begin
       casez(LocAddr)    // decode which register is addressed
                                  
        'hC:     CntrReg           <=  D_IN[`REG_WIDTH-1:0];
        'h10:    WriteOnlyCntrReg  <=  D_IN[`REG_WIDTH-1:0];
        'h20:    HTRsubmodN        <=  D_IN[`REG_WIDTH-1:0];     
        'h28:    ProgInputDelay    <=  D_IN[`REG_WIDTH-1:0];
        'h58:    PipeLength        <=  D_IN[`REG_WIDTH-1:0];
        'h60:    QieChMask[11:0]   <=  D_IN[`REG_WIDTH-1:0];
        'h64:    QieChMask[23:12]  <=  D_IN[`REG_WIDTH-1:0];
        'hB8:    MaxBCN            <=  D_IN[`REG_WIDTH-1:0];
        'hBC:    SetBCN            <=  D_IN[`REG_WIDTH-1:0];
        'hC4: begin
              NPS  <=  D_IN[12:8];
              NS   <=  D_IN[4:0];
              end
 //       'hF0:    histFiber1        <=  D_IN[2:0];      RESERVED
 //       'hF4:    histFiber2        <=  D_IN[2:0];

        'h150:    Threshold1        <=  D_IN[17:0];
        'h154:    Threshold2        <=  D_IN[17:0];
        'h158:    Threshold3        <=  D_IN[17:0];
        'h15C:    Threshold4        <=  D_IN[17:0]; 
        'h160:    Threshold5        <=  D_IN[17:0];
        'h164:    Threshold6        <=  D_IN[17:0];
        'h168:    Threshold7        <=  D_IN[17:0];
        'h16C:    Threshold8        <=  D_IN[17:0];    
        'h170:    Threshold9        <=  D_IN[17:0];
        'h174:    Threshold10        <=  D_IN[17:0];
        'h178:    Threshold11        <=  D_IN[17:0];
        'h17C:    Threshold12        <=  D_IN[17:0]; 
        'h180:    Threshold13        <=  D_IN[17:0];
        'h184:    Threshold14        <=  D_IN[17:0];
        'h188:    Threshold15        <=  D_IN[17:0];
        'h18C:    Threshold16        <=  D_IN[17:0]; 
        'h190:    Threshold17        <=  D_IN[17:0];
        'h194:    Threshold18        <=  D_IN[17:0];
        'h198:    Threshold19        <=  D_IN[17:0];
        'h19C:    Threshold20        <=  D_IN[17:0]; 
        'h1A0:    Threshold21        <=  D_IN[17:0];
        'h1A4:    Threshold22        <=  D_IN[17:0];
        'h1A8:    Threshold23        <=  D_IN[17:0];
        'h1AC:    Threshold24        <=  D_IN[17:0];
		  'h1B0:    ETlow              <=  D_IN[9:0]; 
		  'h1B4:    EThigh             <=  D_IN[9:0]; 
		  'h1BC:    ETch_mux           <=  D_IN[4:0];

      // TriggerPath sub-space
        'b01000_0001_0000_0???_????:   begin   // InLUT1 base addr = 81000
                InLUT_WE  <= 24'b000000000000000000000001;
                InLUT_Mode <= 1'b1;
                end
        'b01000_0010_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000000000000010;
                InLUT_Mode <= 1'b1;
                end
        'b01000_0011_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000000000000100;
                InLUT_Mode <= 1'b1;
                end
        'b01000_0100_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000000000001000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_0101_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000000000010000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_0110_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000000000100000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_0111_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000000001000000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_1000_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000000010000000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_1001_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000000100000000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_1010_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000001000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_1011_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000010000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_1100_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000000100000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_1101_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000001000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_1110_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000010000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01000_1111_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000000100000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01001_0000_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000001000000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01001_0001_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000010000000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01001_0010_0000_0???_????:   begin
                InLUT_WE  <= 24'b000000100000000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01001_0011_0000_0???_????:   begin
                InLUT_WE  <= 24'b000001000000000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01001_0100_0000_0???_????:   begin
                InLUT_WE  <= 24'b000010000000000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01001_0101_0000_0???_????:   begin
                InLUT_WE  <= 24'b000100000000000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01001_0110_0000_0???_????:   begin
                InLUT_WE  <= 24'b001000000000000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01001_0111_0000_0???_????:   begin
                InLUT_WE  <= 24'b010000000000000000000000;
                InLUT_Mode <= 1'b1;
                end
        'b01001_1000_0000_0???_????:   begin
                InLUT_WE  <= 24'b100000000000000000000000;
                InLUT_Mode <= 1'b1;
                end

        'b01001_1001_00??_????_????:   begin   // OutLUT1 base addr = 99000
                OutLUT_WE  <= 24'b000000000000000000000001;
                OutLUT_Mode <= 1'b1;
                end
        'b01001_1010_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000000000000010;
                OutLUT_Mode <= 1'b1;
                end
        'b01001_1011_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000000000000100;
                OutLUT_Mode <= 1'b1;
                end
        'b01001_1100_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000000000001000;
                OutLUT_Mode <= 1'b1;
                end
        'b01001_1101_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000000000010000;
                OutLUT_Mode <= 1'b1;
                end
        'b01001_1110_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000000000100000;
                OutLUT_Mode <= 1'b1;
                end
        'b01001_1111_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000000001000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_0000_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000000010000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_0001_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000000100000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_0010_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000001000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_0011_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000010000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_0100_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000000100000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_0101_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000001000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_0110_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000010000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_0111_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000000100000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_1000_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000001000000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_1001_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000010000000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_1010_00??_????_????:   begin
                OutLUT_WE  <= 24'b000000100000000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_1011_00??_????_????:   begin
                OutLUT_WE  <= 24'b000001000000000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_1100_00??_????_????:   begin
                OutLUT_WE  <= 24'b000010000000000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_1101_00??_????_????:   begin
                OutLUT_WE  <= 24'b000100000000000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_1110_00??_????_????:   begin
                OutLUT_WE  <= 24'b001000000000000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01010_1111_00??_????_????:   begin
                OutLUT_WE  <= 24'b010000000000000000000000;
                OutLUT_Mode <= 1'b1;
                end
        'b01011_0000_00??_????_????:   begin
                OutLUT_WE  <= 24'b100000000000000000000000;
                OutLUT_Mode <= 1'b1;
                end
      // Testing sub-space
        'b11000_0000_00??_????_????:  PatternsWrEn <= 8'b00000001; //patternRAM1 space
        'b11000_0001_00??_????_????:  PatternsWrEn <= 8'b00000010; //patternRAM2 space
        'b11000_0010_00??_????_????:  PatternsWrEn <= 8'b00000100; //patternRAM3 space
        'b11000_0011_00??_????_????:  PatternsWrEn <= 8'b00001000; //patternRAM4 space
        'b11000_0100_00??_????_????:  PatternsWrEn <= 8'b00010000; //patternRAM5 space
        'b11000_0101_00??_????_????:  PatternsWrEn <= 8'b00100000; //patternRAM6 space
        'b11000_0110_00??_????_????:  PatternsWrEn <= 8'b01000000; //patternRAM7 space
        'b11000_0111_00??_????_????:  PatternsWrEn <= 8'b10000000; //patternRAM8 space
                             
                  default:       ;
                  // if there is no such register or is read-only 
                  // no action (will go anyway to END_RD_OR_WR)
            endcase    //    (LocAddr)
             LB_STATE  <=  END_OF_WRITE;          
          end        //  end of MASTER_WRITE case          
    END_OF_WRITE:       begin    
          WriteOnlyCntrReg <=  0;   // self-reset to generate 1-tick pulses
          if(LocCS_Del_1==1)  // level-sensitive to LocCS_b (may avoid lock)
            LB_STATE    <=  END_CYCLE;
          if (NS == 0)  NS  <= 'd20; // safe-sex practices...
          if (NPS > NS) NPS <= 0;
    end
    //  *********************************************
    //    ********** R E A D  operation **********
    MASTER_READ:      begin
          InLUT_WE      <= 0;       InLUT_Mode    <= 1'b1;
          OutLUT_WE     <= 0;       OutLUT_Mode   <= 1'b1;
          PatternsWrEn  <= 0;
       casez(LocAddr)                                    
              // FIFO section: generate Fifo_RD pulses,
              //  the data will be available on the next cycle
        'h34:   InSpyFIFO1_RD      <=  1'b1;      
        'h38:   InSpyFIFO2_RD      <=  1'b1;      
        'h3C:   InSpyFIFO3_RD      <=  1'b1;      
        'h40:   InSpyFIFO4_RD      <=  1'b1;      
        'h44:   InSpyFIFO5_RD      <=  1'b1;      
        'h48:   InSpyFIFO6_RD      <=  1'b1;      
        'h4C:   InSpyFIFO7_RD      <=  1'b1;      
        'h50:   InSpyFIFO8_RD      <=  1'b1;      
        'h54:   L2OutSpyFIFO_RD    <=  1'b1;      
        'hD0:   SnapRdEn[1]        <=  1'b1;      
        'hD4:   SnapRdEn[2]        <=  1'b1;      
        'hD8:   SnapRdEn[3]        <=  1'b1;      
        'hDC:   SnapRdEn[4]        <=  1'b1;      
        'hE0:   SnapRdEn[5]        <=  1'b1;      
        'hE4:   SnapRdEn[6]        <=  1'b1;      
        'hE8:   SnapRdEn[7]        <=  1'b1;      
        'hEC:   SnapRdEn[8]        <=  1'b1;
		  'h1B8:  ETSpyFIFO_RD        <=  1'b1;    
        default:       ;
       endcase   //  (LocAddr)
       LB_STATE  <=  END_OF_READ;          
      end      //  MASTER_READ  case
            
    END_OF_READ:  begin                   
          // release Fifo Read Enable            
          InSpyFIFO1_RD    <=  0;    InSpyFIFO2_RD    <=  0;
          InSpyFIFO3_RD    <=  0;    InSpyFIFO4_RD    <=  0;
          InSpyFIFO5_RD    <=  0;    InSpyFIFO6_RD    <=  0;  
          InSpyFIFO7_RD    <=  0;    InSpyFIFO8_RD    <=  0;  
          SnapRdEn[8:1]    <=  0;    L2OutSpyFIFO_RD  <=  0;
			 ETSpyFIFO_RD     <=  0;

       casez(LocAddr)  // here we actually latch the data outputs
       
        'h8:   D_OUT <=  {3'b0, LastTTCcommand, SD};
        'hC:   D_OUT <=  CntrReg;
        'h14:  D_OUT <=  StatusReg;
        'h18:  D_OUT <=  FIFO_StatReg;
        'h1C:  D_OUT <=  FE_StatReg;
        'h20:  D_OUT <=  HTRsubmodN;
        'h24:  begin          
                  D_OUT[18:16] <=  `FPGAversionType;
                  D_OUT[15:0]  <=  `FPGAversionN;
               end
        'h28:  D_OUT <=  ProgInputDelay;
                        
        'h34:  D_OUT <=  {InSpyFIFO_empty[1], InSpyFIFO1_out};
        'h38:  D_OUT <=  {InSpyFIFO_empty[2], InSpyFIFO2_out};
        'h3C:  D_OUT <=  {InSpyFIFO_empty[3], InSpyFIFO3_out};
        'h40:  D_OUT <=  {InSpyFIFO_empty[4], InSpyFIFO4_out};
        'h44:  D_OUT <=  {InSpyFIFO_empty[5], InSpyFIFO5_out};
        'h48:  D_OUT <=  {InSpyFIFO_empty[6], InSpyFIFO6_out};
        'h4C:  D_OUT <=  {InSpyFIFO_empty[7], InSpyFIFO7_out};
        'h50:  D_OUT <=  {InSpyFIFO_empty[8], InSpyFIFO8_out};
        'h54:  D_OUT <=  {L2OutSpyFIFOempty,  L2OutSpyFIFO_out};
        'h58:  D_OUT <=  PipeLength;
        'h60:  D_OUT <=  {7'b0, QieChMask[11:0]};
        'h64:  D_OUT <=  {7'b0, QieChMask[23:12]};
        'h6C:  D_OUT <=  {3'b0, TTC_DbErrCnt[7:0], TTC_SinErrCnt[7:0]};
        'h70:  D_OUT <=  {3'b0, ErrAcc1};
        'h74:  D_OUT <=  {3'b0, ErrAcc2};
        'h78:  D_OUT <=  {3'b0, ErrAcc3};
        'h7C:  D_OUT <=  {3'b0, ErrAcc4};
        'h80:  D_OUT <=  {3'b0, ErrAcc5};
        'h84:  D_OUT <=  {3'b0, ErrAcc6};
        'h88:  D_OUT <=  {3'b0, ErrAcc7};
        'h8C:  D_OUT <=  {3'b0, ErrAcc8};
        'h90:  D_OUT <=  'hFFFFF;  //  this reg is for testing
        'h94:  D_OUT <=  'h0;      //  this reg is for testing

        'h98:  D_OUT <=  {7'b0, BCNofFib1BZero_r};
        'h9C:  D_OUT <=  {7'b0, BCNofFib2BZero_r};
        'hA0:  D_OUT <=  {7'b0, BCNofFib3BZero_r};
        'hA4:  D_OUT <=  {7'b0, BCNofFib4BZero_r};
        'hA8:  D_OUT <=  {7'b0, BCNofFib5BZero_r};
        'hAC:  D_OUT <=  {7'b0, BCNofFib6BZero_r};
        'hB0:  D_OUT <=  {7'b0, BCNofFib7BZero_r};
        'hB4:  D_OUT <=  {7'b0, BCNofFib8BZero_r};
        'hB8:  D_OUT <=  {7'b0, MaxBCN};
        'hBC:  D_OUT <=  {7'b0, SetBCN};
        'hC4:  D_OUT <=  {6'b0, NPS, 3'b0, NS};
        'hC8:  D_OUT	<=	 {7'b0,  CTrEvN[11:0]};
        'hCC:	D_OUT	<=	 {7'b0,  CTrEvN[`EVN_WIDTH-1:12]};
        'hD0:  D_OUT <=  SnapFIFO1;
        'hD4:  D_OUT <=  SnapFIFO2;
        'hD8:  D_OUT <=  SnapFIFO3;
        'hDC:  D_OUT <=  SnapFIFO4;
        'hE0:  D_OUT <=  SnapFIFO5;
        'hE4:  D_OUT <=  SnapFIFO6;
        'hE8:  D_OUT <=  SnapFIFO7;
        'hEC:  D_OUT <=  SnapFIFO8;
 //       'hF0:	D_OUT	<=	 {16'b0, histFiber1};      RESERVED
 //       'hF4:	D_OUT	<=	 {16'b0, histFiber2};
        'hF8:  D_OUT	<=	 {7'b0,  EvN[11:0]};
        'hFC:	D_OUT	<=	 {7'b0,  EvN[`EVN_WIDTH-1:12]};

        'h100: D_OUT <= TTC_test0;
        'h104: D_OUT <= TTC_test1;
        'h108: D_OUT <= TTC_test2;
        'h10C: D_OUT <= TTC_test3;
        'h110: D_OUT <= TTC_test4;
        'h114: D_OUT <= TTC_test5;
        'h118: D_OUT <= TTC_test6;
        'h11C: D_OUT <= TTC_test7;
        'h120: D_OUT <= TTC_test8;
        'h124: D_OUT <= TTC_test9;
        'h128: D_OUT <= TTC_testA;
        'h12C: D_OUT <= TTC_testB;
        'h130: D_OUT <= TTC_testC;
        'h134: D_OUT <= TTC_testD;
        'h138: D_OUT <= TTC_testE;
        'h13C: D_OUT <= TTC_testF;

        'h140: D_OUT <= {11'b0, LW};
        'h144: D_OUT <= {3'b0, AFifoEmpty, AFifoFull};
        'h148: D_OUT <= {3'b0, AFifo1Status, AFifo2Status, AFifo3Status, AFifo4Status};
        'h14C: D_OUT <= {3'b0, AFifo5Status, AFifo6Status, AFifo7Status, AFifo8Status};

        'h150:    D_OUT <= Threshold1;
        'h154:    D_OUT <= Threshold2;
        'h158:    D_OUT <= Threshold3;
        'h15C:    D_OUT <= Threshold4; 
        'h160:    D_OUT <= Threshold5;
        'h164:    D_OUT <= Threshold6;
        'h168:    D_OUT <= Threshold7;
        'h16C:    D_OUT <= Threshold8;    
        'h170:    D_OUT <= Threshold9;
        'h174:    D_OUT <= Threshold10;
        'h178:    D_OUT <= Threshold11;
        'h17C:    D_OUT <= Threshold12; 
        'h180:    D_OUT <= Threshold13;
        'h184:    D_OUT <= Threshold14;
        'h188:    D_OUT <= Threshold15;
        'h18C:    D_OUT <= Threshold16; 
        'h190:    D_OUT <= Threshold17;
        'h194:    D_OUT <= Threshold18;
        'h198:    D_OUT <= Threshold19;
        'h19C:    D_OUT <= Threshold20; 
        'h1A0:    D_OUT <= Threshold21;
        'h1A4:    D_OUT <= Threshold22;
        'h1A8:    D_OUT <= Threshold23;
        'h1AC:    D_OUT <= Threshold24; 
		  'h1B0:    D_OUT <= ETlow;
		  'h1B4:    D_OUT <= EThigh; 
		  'h1B8:    D_OUT <= {ETSpy_empty,ETSpy_full,2'b00,ETSpy};
		  'h1BC:    D_OUT <= ETch_mux;

      // TriggerPath sub-space
        'b01000_0001_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut1};
        'b01000_0010_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut2};
        'b01000_0011_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut3};
        'b01000_0100_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut4};
        'b01000_0101_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut5};
        'b01000_0110_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut6};
        'b01000_0111_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut7};
        'b01000_1000_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut8};
        'b01000_1001_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut9};
        'b01000_1010_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut10};
        'b01000_1011_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut11};
        'b01000_1100_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut12};
        'b01000_1101_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut13};
        'b01000_1110_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut14};
        'b01000_1111_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut15};
        'b01001_0000_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut16};
        'b01001_0001_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut17};
        'b01001_0010_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut18};
        'b01001_0011_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut19};
        'b01001_0100_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut20};
        'b01001_0101_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut21};
        'b01001_0110_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut22};
        'b01001_0111_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut23};
        'b01001_1000_0000_0???_????:  D_OUT <= {8'b0, InLUT_DOut24};

        'b01001_1001_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut1};
        'b01001_1010_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut2};
        'b01001_1011_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut3};
        'b01001_1100_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut4};
        'b01001_1101_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut5};
        'b01001_1110_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut6};
        'b01001_1111_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut7};
        'b01010_0000_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut8};
        'b01010_0001_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut9};
        'b01010_0010_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut10};
        'b01010_0011_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut11};
        'b01010_0100_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut12};
        'b01010_0101_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut13};
        'b01010_0110_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut14};
        'b01010_0111_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut15};
        'b01010_1000_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut16};
        'b01010_1001_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut17};
        'b01010_1010_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut18};
        'b01010_1011_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut19};
        'b01010_1100_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut20};
        'b01010_1101_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut21};
        'b01010_1110_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut22};
        'b01010_1111_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut23};
        'b01011_0000_00??_????_????:  D_OUT <= {11'b0, OutLUT_DOut24};
      // Testing sub-space
        'b11000_0000_00??_????_????:  D_OUT <= {1'b0, Pattern1Back}; //patternRAM1 space
        'b11000_0001_00??_????_????:  D_OUT <= {1'b0, Pattern2Back}; //patternRAM2 space
        'b11000_0010_00??_????_????:  D_OUT <= {1'b0, Pattern3Back}; //patternRAM3 space
        'b11000_0011_00??_????_????:  D_OUT <= {1'b0, Pattern4Back}; //patternRAM4 space
        'b11000_0100_00??_????_????:  D_OUT <= {1'b0, Pattern5Back}; //patternRAM5 space
        'b11000_0101_00??_????_????:  D_OUT <= {1'b0, Pattern6Back}; //patternRAM6 space
        'b11000_0110_00??_????_????:  D_OUT <= {1'b0, Pattern7Back}; //patternRAM7 space
        'b11000_0111_00??_????_????:  D_OUT <= {1'b0, Pattern8Back}; //patternRAM8 space
                           
        default:  D_OUT <=  'H DEAD;
            // if there is no such read-register answer: DEAD 
       endcase              
       if(LocCS_Del_1==1)  // level-sensitive to LocCS_b (may avoid lock)
            LB_STATE    <=  END_CYCLE;
    end   // END_OF_READ 
          
    END_CYCLE:  
       begin         
             InLUT_Mode    <= 0;        OutLUT_Mode   <= 0;  
             InLUT_WE      <= 0;        OutLUT_WE     <= 0;
             PatternsWrEn  <= 0;        LB_STATE      <= IDLE;
       end
    
    default: LB_STATE    <=  IDLE;

    endcase    //  end the case related to the LB_STATE variable
   end      // else posedge clk
   end        // always
endmodule